# 4. Revert exposing SelectableItem publicly

Date: 2021-07-05

## Status

Accepted

Supercedes [2. Lists of generic items encapsulated with a selection status](0002-lists-of-generic-items-encapsulated-with-a-selection-status.md)

## Context

Originally, the `SelectableList<E>` interface should be based on `ObservableList<SelectableItem<E>>`, that is the encapsulation of list items with a selection status was exposed publicly.

This functionality seems desirable for several reasons, including the following:

* Requesting and changing the selection status of an element does not require access to the list.
* As a consequence, the list can be easily used in conjunction with `javafx.collections.transformation.FilteredList<E>` to filter the list for [un]selected elements.

In the latter case, the predicate to use for filtering does unfortunately _not_ provide the list index, so an `O(n)` search has to be done for each run of the predicate, if the selection status cannot be determined based on the list item alone (that is, filtering the whole list needs at least `O(n²)` time).

But unfortunately, this approach has also some serious disadvantages, including:

* Implementing a `public SelectableItem::setSelected` method would require the item to know its list index, which would have a serious performance impact (e.g. when inserting or deleting elements in the middle of the list)
* Exposing the `SelectableItem<E>` at all would raise the possibility to mix such items between lists, which had to be checked and prevented at runtime (e.g. by throwing a `IllegalStateException`)
* Since, at the end, it is the intention to handle objects of type `<E>` (and not `SelectableItem<E>`), this would require an ugly duplication of the `List<E>` interface. (It is not possible to have the list implementation to implement the `List<E>` and `List<SelectableItem<E>>` interfaces at the same time, so in addition to the `add(SelectableItem<E>)` method, there would be needed an additional `addItem(E)` method, and so on.)

## Decision

The decision to expose the `SelectableItem<E>` interface publicly, e.g. by using an `ObservableList<SelectableItem<E>>` type, has been reverted.

Instead, access to the selection status of list items is possible through indexed access to the `SelectableList<E>` only.

## Consequences

Avoiding to expose the `SelectableItem<E>` interface will result in a simpler and cleaner interface to the `SelectableList<E>` and in a better performance of the data structure.

The disadvantage relating to `javafx.collections.transformation.FilteredList<E>` seems to be acceptable.  
(In case this would really be needed, a `FilteredList<E>` version could be created which provides index information to the filter predicate.)

