/* Created on Oct 21, 2020 */
package de.bernd_michaely.common.desktop.fx.collections.selection;

import de.bernd_michaely.common.desktop.fx.collections.selection.SelectionChangeListener.SelectionChange;
import java.util.function.IntPredicate;
import java.util.stream.IntStream;
import java.util.stream.Stream;
import javafx.beans.property.ReadOnlyBooleanProperty;
import javafx.beans.property.ReadOnlyBooleanWrapper;
import javafx.beans.property.ReadOnlyIntegerProperty;
import javafx.beans.property.ReadOnlyIntegerWrapper;
import javafx.beans.property.ReadOnlyListProperty;
import javafx.beans.property.ReadOnlyListWrapper;
import javafx.beans.property.ReadOnlyObjectProperty;
import javafx.beans.property.ReadOnlyObjectWrapper;
import javafx.collections.transformation.TransformationList;

import static java.util.Objects.requireNonNull;
import static javafx.beans.binding.Bindings.equal;

/**
 * Default implementation of the ListSelectionHandler interface.
 *
 * @author Bernd Michaely (info@bernd-michaely.de)
 * @param <E> the list element type
 */
class ListSelectionHandlerImpl<E> implements ListSelectionHandler<E>
{
	private final ReadOnlyListWrapper<E> sourceListProperty, transformationListProperty;
	private final ReadOnlyObjectWrapper<ReadOnlyListProperty<E>> headList;
	private final ReadOnlyIntegerWrapper numSelectedProperty;
	private final ReadOnlyBooleanWrapper noneSelectedProperty, allSelectedProperty;

	/**
	 * Creates an instance for a given source list.
	 *
	 * @param sourceList the given source list
	 * @throws NullPointerException if sourceList is null
	 */
	ListSelectionHandlerImpl(SelectableList<E> sourceList)
	{
		this(sourceList, null);
	}

	/**
	 * Creates an instance for a given source list and transformation list.
	 *
	 * @param sourceList         the given source list
	 * @param transformationList a transformation list (or a chain thereof) to be
	 *                           applied to the source list (may be null)
	 * @throws NullPointerException     if sourceList is null
	 * @throws IllegalArgumentException if transformationList is not null and
	 *                                  sourceList is not in the transformation
	 *                                  chain
	 */
	ListSelectionHandlerImpl(SelectableList<E> sourceList, TransformationList<E, E> transformationList)
	{
		this.sourceListProperty = new ReadOnlyListWrapper<>(
			requireNonNull(sourceList, getClass().getName() + " : source list is null"));
		if (transformationList != null && !transformationList.isInTransformationChain(sourceList))
		{
			throw new IllegalArgumentException(getClass().getName() +
				" : source list is not in transformation chain");
		}
		this.transformationListProperty = new ReadOnlyListWrapper<>(transformationList);
		this.headList = new ReadOnlyObjectWrapper<>(transformationList != null ?
			this.transformationListProperty : this.sourceListProperty);
		this.numSelectedProperty = new ReadOnlyIntegerWrapper();
		getSourceList().addSelectionListener(this::handleSelectionChange, isTransformed());
		this.noneSelectedProperty = new ReadOnlyBooleanWrapper();
		noneSelectedProperty.bind(equal(numSelectedProperty, 0));
		this.allSelectedProperty = new ReadOnlyBooleanWrapper();
		allSelectedProperty.bind(equal(numSelectedProperty, getHeadList().sizeProperty()));
	}

	@Override
	public SelectableList<E> getSourceList()
	{
		return (SelectableList<E>) sourceListProperty.get();
	}

	@Override
	@SuppressWarnings("unchecked") // OK because set final in constructor
	public TransformationList<E, E> getTransformationList()
	{
		return (TransformationList<E, E>) transformationListProperty.get();
	}

	@Override
	public ReadOnlyObjectProperty<ReadOnlyListProperty<E>> headListProperty()
	{
		return headList.getReadOnlyProperty();
	}

	private void handleSelectionChange(SelectionChange selectionChange)
	{
		if (isTransformed())
		{
			final int sourceIndex = TransformationListUtil.getViewIndexFor(
				getSourceList(), getTransformationList(), selectionChange.getFrom());
			if (sourceIndex >= 0)
			{
				final var selectionChangeType = selectionChange.getSelectionChangeType();
				if (selectionChangeType != null)
				{
					switch (selectionChangeType)
					{
						case SINGLE_INCREMENT ->
							numSelectedProperty.set(numSelectedProperty.get() + 1);
						case SINGLE_DECREMENT ->
							numSelectedProperty.set(numSelectedProperty.get() - 1);
						default -> numSelectedProperty.set((int) streamSelected().count());
					}
				}
			}
		}
		else
		{
			numSelectedProperty.set(getSourceList().getNumSelected());
		}
	}

	@Override
	public ReadOnlyBooleanProperty allSelectedProperty()
	{
		return allSelectedProperty.getReadOnlyProperty();
	}

	@Override
	public ReadOnlyBooleanProperty noneSelectedProperty()
	{
		return noneSelectedProperty.getReadOnlyProperty();
	}

	@Override
	public ReadOnlyIntegerProperty numSelectedProperty()
	{
		return numSelectedProperty.getReadOnlyProperty();
	}

	@Override
	public ReadOnlyIntegerProperty sizeProperty()
	{
		return getHeadList().sizeProperty();
	}

	@Override
	public ReadOnlyBooleanProperty emptyProperty()
	{
		return getHeadList().emptyProperty();
	}

	@Override
	public void selectRange(int from, int to, Action action)
	{
		IntStream.range(from, to)
			.map(this::getSourceIndex)
			.forEach(index -> getSourceList().select(index, action));
	}

	@Override
	public int getSourceIndex(int index)
	{
		return getTransformationList() != null ?
			getTransformationList().getSourceIndexFor(getSourceList(), index) : index;
	}

	@Override
	public E get(int index)
	{
		return getHeadList().get(index);
	}

	/**
	 * Returns true, iff the item at the given index is selected.
	 *
	 * @param index the given index, including any transformations
	 * @return true, iff the item at the given index is selected
	 */
	@Override
	public boolean isSelected(int index)
	{
		return getSourceList().isSelected(getSourceIndex(index));
	}

	@Override
	public Stream<E> streamSelected()
	{
		return streamFiltered(this::isSelected);
	}

	@Override
	public Stream<E> streamUnselected()
	{
		final IntPredicate selected = this::isSelected;
		return streamFiltered(selected.negate());
	}

	private Stream<E> streamFiltered(IntPredicate predicate)
	{
		return IntStream.range(0, size()).filter(predicate).mapToObj(this::get);
	}

	/**
	 * {@inheritDoc}
	 *
	 * @return a String representation of the items of the head list with their
	 *         selection status
	 */
	@Override
	public String toString()
	{
		final SelectableList<E> list;
		if (isTransformed())
		{
			list = SelectableListFactory.selectableList();
			for (int i = 0; i < size(); i++)
			{
				list.add(get(i));
				if (isSelected(i))
				{
					list.setSelected(i, true);
				}
			}
		}
		else
		{
			list = getSourceList();
		}
		return list.toString();
	}
}
