/* Created on Oct 16, 2020 */
package de.bernd_michaely.common.desktop.fx.collections.selection;

import java.lang.ref.WeakReference;
import java.util.ArrayList;
import java.util.Collection;
import java.util.List;
import java.util.Objects;
import java.util.RandomAccess;
import javafx.collections.ModifiableObservableListBase;

/**
 * Default implementation of the SelectableList interface. This package local
 * implementation should be instantiated by the {@link CommonFXCollections}
 * factory class only.
 *
 * @param <E> the list element type
 * @author Bernd Michaely (info@bernd-michaely.de)
 */
class SelectableListImpl<E> extends ModifiableObservableListBase<E>
	implements SelectableList<E>, RandomAccess
{
	/**
	 * Class to encapsulate a list item with a selection state.
	 *
	 * @param <E> the list element type
	 */
	private static class SelectableItem<E>
	{
		private E item;
		private boolean selected;

		private SelectableItem(E item)
		{
			this.item = item;
		}

		@Override
		public boolean equals(Object object)
		{
			return (object instanceof SelectableItem other) ?
				Objects.equals(this.item, other.item) : false;
		}

		@Override
		public int hashCode()
		{
			return Objects.hashCode(item);
		}

//		private static final char CHAR_SELECTED = '✓';
//		private static final char CHAR_UNSELECTED = '✕';
		private static final char CHAR_SELECTED = '+';
		private static final char CHAR_UNSELECTED = '-';

		@Override
		public String toString()
		{
			return "»" + item + "«" + (selected ? CHAR_SELECTED : CHAR_UNSELECTED);
		}
	}

	private final List<SelectableItem<E>> delegate;
	private final SelectionControl<E> selectionControl;

	SelectableListImpl()
	{
		this.delegate = new ArrayList<>();
		this.selectionControl = new SelectionControl<>(new WeakReference<>(this));
	}

	SelectableListImpl(Collection<? extends E> src)
	{
		this();
		addAll(src);
	}

	@Override
	public void beginSelectionChange()
	{
		selectionControl.beginSelectionChange();
	}

	@Override
	public void endSelectionChange()
	{
		selectionControl.endSelectionChange();
	}

	@Override
	public E get(int index)
	{
		return delegate.get(index).item;
	}

	@Override
	protected void doAdd(int index, E item)
	{
		delegate.add(index, new SelectableItem<>(item));
	}

	@Override
	protected E doRemove(int index)
	{
		final SelectableItem<E> selectableItem = delegate.remove(index);
		if (selectableItem.selected)
		{
			selectionControl.decrementSelectionCounter(index);
		}
		return selectableItem.item;
	}

	@Override
	protected E doSet(int index, E item)
	{
		final SelectableItem<E> selectableItem = delegate.get(index);
		final E oldValue = selectableItem.item;
		selectableItem.item = item;
		return oldValue;
	}

	@Override
	public int size()
	{
		return delegate.size();
	}

	@Override
	public boolean isSelected(int index)
	{
		return delegate.get(index).selected;
	}

	@Override
	public void selectRange(int from, int to, Action action)
	{
		if (from < to && action != null)
		{
			beginSelectionChange();
			try
			{
				for (int i = from; i < to; i++)
				{
					final SelectableItem<E> item = delegate.get(i);
					final boolean oldValue = item.selected;
					final boolean newValue = switch (action)
					{
						case SELECTION_SET ->
							true;
						case SELECTION_UNSET ->
							false;
						case SELECTION_TOGGLE ->
							!oldValue;
					};
					if (oldValue != newValue)
					{
						item.selected = newValue;
						if (newValue)
						{
							selectionControl.incrementSelectionCounter(i);
						}
						else
						{
							selectionControl.decrementSelectionCounter(i);
						}
					}
				}
			}
			finally
			{
				endSelectionChange();
			}
		}
	}

	@Override
	public int getNumSelected()
	{
		return selectionControl.getSelectionCounter();
	}

	@Override
	public void addSelectionListener(SelectionChangeListener<? super E> selectionChangeListener,
		boolean requestSingleChangeEvents)
	{
		selectionControl.addSelectionListener(selectionChangeListener, requestSingleChangeEvents);
	}

	@Override
	public boolean removeSelectionListener(SelectionChangeListener<? super E> selectionChangeListener)
	{
		return selectionControl.removeSelectionListener(selectionChangeListener);
	}

	@Override
	public String toString()
	{
		return delegate.toString();
	}
}
