/* Created on Nov 29, 2020 */
package de.bernd_michaely.common.desktop.fx.collections.selection;

import de.bernd_michaely.common.desktop.fx.collections.selection.SelectionChangeListener.SelectionChange;
import java.lang.ref.WeakReference;
import java.util.IdentityHashMap;

import static java.util.Objects.requireNonNull;
import static java.util.function.Predicate.not;

/**
 * Internal class to control the current count of selected list items and the
 * corresponding event handling. All {@link SelectableItem}s and the
 * {@link SelectableList} itself keep a reference to one instance of this class
 * to communicate through it. The {@link SelectableList} is accessed through a
 * {@link  WeakReference} to avoid memory leaks. The counting is done in
 * {@code O(1)} time.
 *
 * @author Bernd Michaely (info@bernd-michaely.de)
 */
class SelectionControl<E>
{
	private final WeakReference<SelectableList<E>> selectableList;
	private final IdentityHashMap<SelectionChangeListener<? super E>, Boolean> selectionChangeListeners;
	private int stackCounter;
	private int selectionCounter;
	private SelectionChange<E> selectionChange;

	SelectionControl(WeakReference<SelectableList<E>> selectableList)
	{
		this.selectableList = requireNonNull(selectableList, "SelectableList is null");
		this.selectionChangeListeners = new IdentityHashMap<>();
	}

	int getSelectionCounter()
	{
		return selectionCounter;
	}

	void incrementSelectionCounter(int indexChanged)
	{
		beginSelectionChange();
		try
		{
			selectionCounter++;
			selectionChange.addIndex(indexChanged, true);
			notifySingleChangeRequestListeners(indexChanged, true);
		}
		finally
		{
			endSelectionChange();
		}
	}

	void decrementSelectionCounter(int indexChanged)
	{
		if (selectionCounter > 0)
		{
			beginSelectionChange();
			try
			{
				selectionCounter--;
				selectionChange.addIndex(indexChanged, false);
				notifySingleChangeRequestListeners(indexChanged, false);
			}
			finally
			{
				endSelectionChange();
			}
		}
		else
		{
			throw new IllegalStateException(getClass().getName() +
				"#decrementSelectionCounter : invalid call");
		}
	}

	private void notifySingleChangeRequestListeners(int index, boolean selected)
	{
		selectionChangeListeners.keySet().stream()
			.filter(selectionChangeListeners::get)
			.forEach(listener ->
			{
				final SelectionChange<E> change = new SelectionChange<>(selectableList);
				change.addIndex(index, selected);
				listener.onSelectionChanged(change);
			});
	}

	void beginSelectionChange()
	{
		if (stackCounter < 0)
		{
			throw new IllegalStateException(getClass().getName() +
				"#beginSelectionChange : invalid stack counter");
		}
		if (++stackCounter == 1)
		{
			selectionChange = new SelectionChange<>(selectableList);
		}
	}

	void endSelectionChange()
	{
		if (--stackCounter >= 0)
		{
			if (stackCounter == 0)
			{
				try
				{
					if (selectionChange != null)
					{
						selectionChangeListeners.keySet().stream()
							.filter(not(selectionChangeListeners::get))
							.forEach(listener -> listener.onSelectionChanged(selectionChange));
					}
					else
					{
						throw new IllegalStateException("Invalid call of notifyListeners()");
					}
				}
				finally
				{
					selectionChange = null;
				}
			}
		}
		else
		{
			throw new IllegalStateException("Invalid call of endSelectionChange()");
		}
	}

	/**
	 * Adds the given listener.
	 *
	 * @param selectionChangeListener the listener to add
	 * @throws NullPointerException if the given listener is null
	 */
	void addSelectionListener(SelectionChangeListener<? super E> selectionChangeListener,
		boolean requestSingleChangeEvents)
	{
		selectionChangeListeners.put(requireNonNull(selectionChangeListener), requestSingleChangeEvents);
	}

	/**
	 * Removes the given listener.
	 *
	 * @param selectionChangeListener the listener to remove
	 * @return true, iff the listener was present before
	 * @throws NullPointerException if the given listener is null
	 */
	boolean removeSelectionListener(SelectionChangeListener<? super E> selectionChangeListener)
	{
		return selectionChangeListeners.remove(requireNonNull(selectionChangeListener)) != null;
	}
}
